import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';

import 'core.dart';
import 'stream_listener.dart';

typedef TruckWidgetListener<B, S> = void Function(BuildContext context, B truck, S state);

class TruckListener<B extends EventStreamableSource<S>, S> extends TruckListenerBase<B, S>
    with LuxListenerSingleChildWidget {
  /// {@macro truck_listener}
  TruckListener({
    Key? key,
    required TruckWidgetListener<B, S> listener,
    B? truck,
    StreamListenerCondition<S>? listenWhen,
    Widget? child,
  }) : super(
          key: key,
          listener: listener,
          truck: truck,
          child: child,
          listenWhen: listenWhen,
        );
}

class TruckListenerBase<B extends EventStreamableSource<S>, S> extends SingleChildStatefulWidget {
  /// {@macro truck_listener_base}
  const TruckListenerBase({
    Key? key,
    required this.listener,
    this.truck,
    this.listenWhen,
    this.child,
  }) : super(key: key, child: child);

  final B? truck;

  final Widget? child;

  final TruckWidgetListener<B, S> listener;

  final StreamListenerCondition<S>? listenWhen;

  @override
  SingleChildState<TruckListenerBase<B, S>> createState() => _TruckListenerBaseState<B, S>();
}

class _TruckListenerBaseState<B extends EventStreamableSource<S>, S> extends SingleChildState<TruckListenerBase<B, S>> {
  StreamSubscription<S>? _subscription;
  late B _truck;

  @override
  void initState() {
    super.initState();
    _truck = widget.truck ?? context.read<B>();
    _subscribe();
  }

  @override
  void didUpdateWidget(TruckListenerBase<B, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldTruck = oldWidget.truck ?? context.read<B>();
    final currentTruck = widget.truck ?? oldTruck;
    if (oldTruck != currentTruck) {
      if (_subscription != null) {
        _unsubscribe();
        _truck = currentTruck;
      }
      _subscribe();
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final truck = widget.truck ?? context.read<B>();
    if (_truck != truck) {
      if (_subscription != null) {
        _unsubscribe();
        _truck = truck;
      }
      _subscribe();
    }
  }

  @override
  Widget buildWithChild(BuildContext context, Widget? child) {
    assert(child != null, '''${widget.runtimeType} used outside of TruckListener must specify a child''');
    if (widget.truck == null) {
      context.select<B, bool>((truck) => identical(_truck, truck));
    }
    return child!;
  }

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  void _subscribe() {
    _subscription = _truck.stream.listen((state) {
      if (widget.listenWhen == null || widget.listenWhen!(state) == true) {
        widget.listener(context, _truck, state);
      }
    });
  }

  void _unsubscribe() {
    _subscription?.cancel();
    _subscription = null;
  }
}
